const zk = require('node-zookeeper-client');
const log4js = require('log4js');
const _ = require('lodash');

class ZkClient {
     _options = {};
     _connected = false;
     _zkClient = undefined;
     _createMode = undefined;
     _logger = undefined;

    constructor(options) {
        this._options = options || {};

        this._logger = log4js.getLogger('ZkClient');

        this._zkClient = zk.createClient(
            `${this._options.host}:${this._options.port}`, 
            { 
                retries: this._options.retries || 5,
                spinDelay: this._options.spinDelay || 1000,
                sessionTimeout: this._options.sessionTimeout || 180000 
            });

        this._createMode = zk.CreateMode;
    }

    async connect() {
        const self = this;

        return new Promise((resolve, reject) => {
            const connected = () => {
                self._logger.debug(`Connected to hosts: ${self._options.host}:${self._options.port}, retires: ${self._options.retries}`);
                self._connected = true;
                resolve(true);
            };

            const disconnected = () => {
                self._logger.debug(`Disconnected from hosts: ${self._options.host}:${self._options.port}`);
                self._connected = false;
            };

            const sessionExpired = () => {
                self._logger.warn('Session expired.');
            }

            self._zkClient.once('connected', connected);
            self._zkClient.once('disconnected', disconnected);
            self._zkClient.once('expired', sessionExpired);

            try {
                self._zkClient.connect();
            } catch(e) {
                self._logger.error(`Error connecting to zookeeper host(s): ${self._options.host}:${self._options.port} - Error: ${e}`);
                reject(e);
            }
        });
    }

    async close() {
        const self = this;

        return new Promise((resolve, reject) => {
            try {
                self._zkClient.close();
                resolve(true);                
            } catch(e) {
                self._logger.error(`Error closing connection to zookeeper host(s): ${self._options.host}:${self._options.port} - Error: ${e}`);
                reject(e);
            }
        });
    }

    async create(znode, data, acls, mode) {
        const self = this;

        const pathElements = znode.split('/').filter(item => item.trim());

        // last element is node name, so discard it
        pathElements.pop();

        // join remaining elements to get fully qualified path
        let parentNode = '/';
        if (pathElements.length > 0)
            parentNode = pathElements.join('/');

        const nodeInfo = {
            parentZnode: parentNode,
            znode: znode,
            name: znode.split('/').pop()
        };

        const existingNode = await self.exists(znode);

        return new Promise((resolve, reject) => {
            if (existingNode) {
                resolve(nodeInfo);
            } else {
                const nodeData = data ? Buffer.from(JSON.stringify(data), 'utf-8') : null;

                self._zkClient.create(znode, nodeData, acls, mode, function(err, path) {
                    if (err) {
                        self._logger.error(`Error creating node: ${znode} - Error: ${err.message}, StackTrace: ${err.stack}`);
                        reject(err.message);
                    }

                    resolve({
                        parentNode: parentNode,
                        znode: path,
                        name: path.split('/').pop()
                    });
                });
            }
        });
    }

    async remove(znode) {
        const self = this;

        return new Promise((resolve, reject) => {
            self._zkClient.remove(znode, -1, (err) => {
                if (err) {
                    self._logger.error(`Error deleting node ${znode} - Error: ${err.message}, StackTrace: ${err.stack}`);
                    return reject(err.message);
                }
    
                return resolve({
                    znode: znode,
                    name: znode.split('/').pop()
                });
            });
        });
    }

    async removeWithRetries(znode, retryCounter) {
        const self = this;

        try {
            const removedNode = await self.remove(znode);

            return removedNode;
        } catch(e) {
            self._logger.error(`Retry number: ${retryCounter} - Error: ${e}`);
            if (retryCounter > 0) {
                retryCounter -= 1;
                return await self.removeWithRetries(znode, retryCounter);
            }

            self._logger.error(`All retries to remove this node failed: ${znode}`);
            throw new Error(`All retries to remove '${znode}' failed.`);
        }
    }

    async exists(znode) {
        const self = this;

        return new Promise((resolve, reject) => {
            self._zkClient.exists(znode, null, (err, stat) => {
                if (err) {
                    self._logger.error(`Error while checking existence of node ${znode}. Error: ${err.message}, StackTrace: ${err.stack}`);
                    reject(err.message);
                } else {
                    resolve(stat ? true : false);
                }
            });
        });
    }

    async getChildren(znode) {
        const self = this;

        return new Promise((resolve, reject) => {
            self._zkClient.getChildren(znode, (err, children) => {
                if (err) {
                    if (!/NO_NODE/i.test(err.name)) {
                        self._logger.error(`Error occured while trying to get children of node ${znode} - Error: ${err.message}, StackTrace: ${err.stack}`);
                        return reject(err.message);
                    }
                }

                return resolve(children);
            });
        });
    }

    async getData(znode) {
        const self = this;

        return new Promise((resolve, reject) => {
            self._zkClient.getData(znode, (err, data) => {
                if (err) {
                    self._logger.error(`Error getting data for node: ${znode} - Error: ${err.message}, StackTrace: ${err.stack}`);
                    return reject(err.message);
                }

                return resolve({
                    znode: znode,
                    name: znode.split('/').pop(),
                    data: data && data.length > 0 ? JSON.parse(data) : null
                });
            });
        });
    }

    async setData(znode,  data) {
        const self = this;

        return new Promise((resolve, reject) => {
            const nodeData = data ? Buffer.from(JSON.stringify(data), 'utf-8') : null;

            self._zkClient.setData(znode, nodeData, (err, stat) => {
                if (err) {
                    self._logger.error(`Error setting data for node: ${znode} - Error: ${err.message}, StackTrace: ${err.stack}`);
                    return reject(err.message);
                }

                return resolve({
                    znode: znode,
                    name: znode.split('/').pop(),
                    stat: stat,
                    data: data
                });
            });
        })
    }

    async safeSetData(znode, data) {
        const self = this;
        let updatedNode;

        try {
            const lock = await self.acquireLock(znode);

            updatedNode = await self.setData(znode, data);

            await self.remove(lock.znode);
        } catch(e) {
            self._logger.error(`Error occured while setting data for this node: ${znode} - Error: ${e}`);
            throw e;
        }

        return updatedNode;
    }

    async acquireLock(znode) {
        var self = this;

        async function _waitForTheLock(lckNode) {
            function _doWeHaveTheLock(children) {
                var sortedLockNodes = _.filter(children, x => /lock-/i.test(x)).sort();
    
                // if we are the first then we have the lock
                if (lckNode.name == _.first(sortedLockNodes)) {
                    return { acquired: true };
                } else {
                    var index =_.findIndex(sortedLockNodes, function(item) {
                        return item == lckNode.name;
                    });
    
                    return {
                        znode: `${znode}/${sortedLockNodes[index - 1]}`,
                        acquired: false
                    };
                }
            }

            async function _checkIfLockIsAcquired() {
                try {
                    const children = await self.getChildren(znode);
                    const lockStatus = _doWeHaveTheLock(children);

                    if (lockStatus.acquired)
                        return lckNode;

                    return await _waitAndCheck(lockStatus.znode);
                } catch(e) {
                    const msg = `Error in _checkIfLockAcquired for ${znode} - Error: ${e}`;
                    self._logger(msg);
                    throw new Error(msg);
                }
            }
    
            async function _waitAndCheck(targetNode) {
                return new Promise((resolve, reject) => {
                    self._zkClient.exists(targetNode, async () => {
                        try {
                            const result = await _checkIfLockIsAcquired();
                            resolve(result);
                        } catch(e) {
                            reject(`Fail to acquire lock on ${znode} - Error: ${e}`);
                        }
                    }, async (err, stat) => {
                        if (err)
                            reject(`Fail to acquire lock on node ${znode} - Error: ${err.message}, StackTrace: ${err.stack}`);
                        else if (!stat) {
                            try {
                                const result = await _checkIfLockIsAcquired();
                                resolve(result);
                            } catch(e) {
                                reject(`Fail to acquire lock on ${znode} - Error: ${e}`);
                            }
                        }
                    });
                });
            }
    
            return await _checkIfLockIsAcquired();
        }
    
        try {
            const children = await self.getChildren(znode);
            if (!_.isEmpty(_.filter(children, x => /lock-/i.test(x.name))))
                throw new Error(`${znode} is already locked.`);

            const lockNode = await self.create(`${znode}/lock-`, null, null, self._createMode.EPHEMERAL_SEQUENTIAL);
            const res = await _waitForTheLock(lockNode);

            return res;
        } catch(e) {
            const msg = `${znode} is already locked - Error: ${e}`;

            self._logger.error(msg);
            throw new Error(msg);
        }
    }

    async releaseLock(lock) {
        if (!lock)
            return null;

        const self = this;
        const targetZnode = _.isObject(lock) ? lock.znode : lock;

        await self.remove(targetZnode);
    }
}

module.exports = ZkClient;